#include <linux/types.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/bitops.h>
#include <linux/proc_fs.h>
#include <asm/spinlock.h>
#include <linux/netfilter_ipv4.h>

#include "ip_nat_proto.h"
#include "ip_nat_rule.h"
#include "ip_nat_binding.h"

#define IP_NAT_TCP_ADD_USER (IP_NAT_SO_MAX+1)
#define IP_NAT_TCP_DELETE_USER (IP_NAT_SO_MAX+2)

#define IP_NAT_TCP_OPENING_TIMEOUT (75*HZ)
#define IP_NAT_TCP_ESTABLISHED_TIMEOUT (6*3600*HZ)
#define IP_NAT_TCP_CLOSED_TIMEOUT (75*HZ)

extern struct ip_nat_proto proto_tcp;

struct ip_nat_tcp_user_port
{
	struct ip_nat_tcp_user_port *next;
	u_int16_t port;
	ip_nat_fn *matchfn;
	ip_nat_fn *relatedfn;
	void (*timerfn)(struct ip_nat_binding *binding);
	unsigned long timeout;
};

enum ip_nat_tcpstate {
	IP_NAT_TCP_OPENING,
	IP_NAT_TCP_ESTABLISHED,
	IP_NAT_TCP_CLOSING,
};

static struct ip_nat_tcp_user_port *userports;

/* Use this to track binding state. */
struct ip_nat_binding_tcp
{
	enum ip_nat_tcpstate state;

	/* Manip we were initialized on. */
	enum ip_nat_manip_type nat_or_rnat;
};

static int tcp_check(const struct ip_nat_rule *rule, unsigned int typenum)
{
	/* Must specify proto == TCP. */
	if (rule->elem.mask.tuplepr.proto != 0xFF
	    || rule->elem.info.tuplepr.proto != IPPROTO_TCP) {
		DEBUGP("tcp_rule_check: mask %u, proto = %u.\n",
		       rule->elem.mask.tuplepr.proto,
		       rule->elem.info.tuplepr.proto);
		return 0;
	}

	if (rule->elem.type[typenum+1]) {
		DEBUGP("tcp_rule_check: must be last.\n");
		return 0;
	}

	return 1;
}

/* We don't care when you destroy bindings. */
#ifdef DEBUGNAT
static void tcp_destroy(const struct ip_nat_binding *binding)
{
	struct ip_nat_binding_tcp *binding_tcp
		= ((void *)binding) + proto_tcp.mappingtype.private_offset;

	if (binding_tcp->state != IP_NAT_TCP_CLOSING)
		DEBUGP("tcp_destroy: destroying in state %s\n",
		       binding_tcp->state == IP_NAT_TCP_ESTABLISHED
		       ? "ESTABLISHED"
		       : (binding_tcp->state == IP_NAT_TCP_OPENING
			  ? "OPENING" : "UNKNOWN"));
}
#endif

static unsigned int process_icmp(struct ip_nat_binding *binding,
				 struct iphdr *iph,
				 u_int16_t len,
				 enum ip_nat_manip_type manip_type)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_binding_tcp *binding_tcp
		= ((void *)binding) + proto_tcp.mappingtype.private_offset;

	if (binding_tcp->state == IP_NAT_TCP_OPENING
	    && hdr->type == ICMP_DEST_UNREACH) {
		mod_timer(&binding->timeout,
			  jiffies + IP_NAT_TCP_CLOSED_TIMEOUT);
		binding_tcp->state = IP_NAT_TCP_CLOSING;
	}
	return NF_ACCEPT;
}

static unsigned int process(struct ip_nat_binding *binding,
			    struct iphdr *iph,
			    u_int16_t len,
			    enum ip_nat_manip_type manip_type)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_binding_tcp *binding_tcp
		= ((void *)binding) + proto_tcp.mappingtype.private_offset;

	switch (binding_tcp->state) {
	case IP_NAT_TCP_OPENING:
		DEBUGP("ip_nat_tcp: IP_NAT_TCP_OPENING.\n");
		if (manip_type != binding_tcp->nat_or_rnat) {
			/* First reply packet: RST -> move to CLOSING,
			 * Other -> move to ESTABLISHED. */
			if (!(ntohs(iph->frag_off) & IP_OFFSET)) {
				if (hdr->rst) {
					mod_timer(&binding->timeout,
						  jiffies
						  + IP_NAT_TCP_CLOSED_TIMEOUT);
					DEBUGP("ip_nat_tcp: RST!\n");
					binding_tcp->state
						= IP_NAT_TCP_CLOSING;
				} else {
					mod_timer(&binding->timeout,
						  jiffies
						  + IP_NAT_TCP_ESTABLISHED_TIMEOUT);
					DEBUGP("ip_nat_tcp: OK!\n");
					binding_tcp->state
						= IP_NAT_TCP_ESTABLISHED;
				}
			}
		} else {
			DEBUGP("ip_nat_tcp: Refresh!\n");
			/* Refresh timer. */
			mod_timer(&binding->timeout,
				  jiffies + IP_NAT_TCP_OPENING_TIMEOUT);
		}
		break;

	case IP_NAT_TCP_CLOSING:
		DEBUGP("ip_nat_tcp: IP_NAT_TCP_CLOSING.\n");
		/* Nothing can change it now; we're doomed. */

		/* Maybe a new connection?  If so, drop and delete
		   this binding to give it a fresh chance. */
		if (!(ntohs(iph->frag_off) & IP_OFFSET)
		    && !hdr->rst && !hdr->fin && !hdr->ack && hdr->syn) {
			printk("Unexpected TCP reconnect on %s (rule from %s)"
			       " %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n",
			       manip_type == IP_NAT_MANIP_SRC ? "SRC" : "DST",
			       binding_tcp->nat_or_rnat == IP_NAT_MANIP_SRC
			       ? "SRC" : "DST",
			       IP_PARTS(iph->saddr), hdr->source,
			       IP_PARTS(iph->daddr), hdr->dest);
			ip_nat_delete_binding_timeout(binding);
			return NF_DROP;
		}
		return NF_ACCEPT;

	case IP_NAT_TCP_ESTABLISHED:
		DEBUGP("ip_nat_tcp: IP_NAT_TCP_ESTABLISHED.\n");
		/* If we see a RST or a FIN, enter expiry cycle. */
		if (!(ntohs(iph->frag_off) & IP_OFFSET)
		    && (hdr->fin || hdr->rst)) {
			mod_timer(&binding->timeout,
				  jiffies + IP_NAT_TCP_CLOSED_TIMEOUT);
			binding_tcp->state = IP_NAT_TCP_CLOSING;
		} else {
			/* Refresh timer */
			mod_timer(&binding->timeout,
				  jiffies + IP_NAT_TCP_ESTABLISHED_TIMEOUT);
		}
		break;
	default:
		DEBUGP("ip_nat_tcp: default %u.\n", binding_tcp->state);
	}

	return NF_ACCEPT;
}

static struct ip_nat_tcp_user_port no_user_port =
{ NULL, 0, &process, &process_icmp,
 &ip_nat_delete_binding_timeout, IP_NAT_TCP_OPENING_TIMEOUT };

static int port_cmp(const void *arg, const void *curr)
{
	return ((struct ip_nat_tcp_user_port *)curr)->port
		== (u_int16_t)(unsigned long)arg;
}

/* Function to call if we get a match */
static unsigned int tcp_rule(const struct iphdr *iph,
			     const struct ip_nat_pktinfo *pkt,
			     const struct ip_nat_rule_element *cause,
			     struct ip_nat_multi_range *mr,
			     unsigned int typenumber,
			     struct ip_nat_binding **binding)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_tcp_user_port *user;
	struct ip_nat_binding_tcp *binding_tcp;

	/* Send to userspace if it's in the user port list. */
	user = modreg_find(&userports, (void *)(int)hdr->dest, &port_cmp);
	if (!user)
		user = &no_user_port;

	dump_tuplepr(__FUNCTION__, &pkt->tuplepr, pkt->hook);
	*binding = ip_nat_alloc_binding(pkt, cause, mr, user->matchfn,
					user->relatedfn, user->timeout,
					&ip_nat_delete_binding_timeout);
	if (!*binding) {
		DEBUGP("tcp-nat: ip_nat_alloc_binding returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("tcp-nat: ip_nat_alloc_binding returned ok.\n");

	binding_tcp = ((void *)*binding) + proto_tcp.mappingtype.private_offset;
	*binding_tcp = ((struct ip_nat_binding_tcp)
			{ IP_NAT_TCP_OPENING, pkt->hook });

	return NF_ACCEPT;
}

static int do_tcp_ctl(int cmd, void *user, unsigned int len)
{
	struct ip_nat_tcp_user_port *portinfo;

	switch (cmd) {
	case IP_NAT_TCP_ADD_USER:
		portinfo = kmalloc(sizeof(*portinfo), GFP_USER);
		if (!portinfo)
			return -ENOMEM;
		else {
			portinfo->port = (u_int16_t)(unsigned long)user;
			modreg_prepend(&userports, portinfo);
		}
		break;

	case IP_NAT_TCP_DELETE_USER:
		portinfo = modreg_find(&userports, user, &port_cmp);

		if (!portinfo)
			return -ENOENT;
		else {
			modreg_delete(&userports, portinfo);
			kfree(portinfo);
		}
	}

	return 0;
}

static int tcp_in_range(const struct ip_nat_tuple_proto *tuplepr,
			enum ip_nat_manip_type manip_type,
			const union ip_nat_proto_manip *min,
			const union ip_nat_proto_manip *max)
{
	u_int16_t port;

	if (manip_type == IP_NAT_MANIP_SRC)
		port = tuplepr->tuple.src.u.tcp.port;
	else
		port = tuplepr->tuple.dst.u.tcp.port;

	return ntohs(port) >= ntohs(min->tcp.port)
		&& ntohs(port) <= ntohs(max->tcp.port);
}

static int tcp_unique_tuple(struct ip_nat_tuple_proto *tuplepr,
			    const struct ip_nat_range *range,
			    unsigned int hook)
{
	static u_int16_t port = 0, *portptr;
	unsigned int range_size, min, i;

	if (HOOK2MANIP(hook) == IP_NAT_MANIP_SRC)
		portptr = &tuplepr->tuple.src.u.tcp.port;
	else
		portptr = &tuplepr->tuple.dst.u.tcp.port;

	/* If no range specified... */
	if (!(range->flags & IP_NAT_RANGE_PROTO_SPECIFIED)) {
		/* If it's dst rewrite, can't change port */
		if (HOOK2MANIP(hook) == IP_NAT_MANIP_DST)
			return 0;

		if (*portptr < 1024) {
			DEBUGP("Privileged Range map.\n");
			min = 1;
			range_size = 1023;
		} else {
			DEBUGP("Unprivileged Range map.\n");
			min = 1024;
			range_size = 65535 - 1024 + 1;
		}
	} else {
		min = ntohs(range->min.tcp.port);
		range_size = ntohs(range->max.tcp.port) - min + 1;
	}

	for (i = 0; i < range_size; i++, port++) {
		*portptr = htons(min + port % range_size);
		if (!ip_nat_used_tuple(tuplepr, hook)) {
			DEBUGP("Allocated port %u.\n", ntohs(*portptr));
			return 1;
		}
		else DEBUGP("port %u already taken.\n", ntohs(*portptr));
	}
	return 0;
}

static int tcp_pkt_to_tuple(const void *datah, size_t datalen,
			    struct ip_nat_tuple *tuple)
{
	const struct tcphdr *hdr = datah;
	if ((char *)&hdr->dest + sizeof(hdr->dest) - (char *)datah > datalen)
		return 0;
	tuple->src.u.tcp.port = hdr->source;
	tuple->dst.u.tcp.port = hdr->dest;

	return 1;
}

static void tcp_manip_pkt(struct iphdr *iph, size_t len,
			  const struct ip_nat_tuple_manip *manip,
			  enum ip_nat_manip_type manip_type)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	u_int32_t oldip;
	u_int16_t *portptr;

	if (manip_type == IP_NAT_MANIP_SRC) {
		/* Get rid of src ip and src pt */
		oldip = iph->saddr;
		portptr = &hdr->source;
	} else {
		/* Get rid of dst ip and dst pt */
		oldip = iph->daddr;
		portptr = &hdr->dest;
	}
	DEBUGP("TCP: IP %u.%u.%u.%u to %u.%u.%u.%u, port %u to %u\n",
	       IP_PARTS(oldip), IP_PARTS(manip->ip),
	       htons(*portptr), htons(manip->u.tcp.port));

	hdr->check = ip_nat_cheat_check(~oldip, manip->ip,
					ip_nat_cheat_check(*portptr ^ 0xFFFF,
							   manip->u.tcp.port,
							   hdr->check));
	*portptr = manip->u.tcp.port;
}

static void tcp_invert_tuple(struct ip_nat_tuple *tuple)
{
	/* Switch source and destination ports */
	u_int16_t tmp;

	tmp = tuple->src.u.tcp.port;
	tuple->src.u.tcp.port = tuple->dst.u.tcp.port;
	tuple->dst.u.tcp.port = tmp;
}

static unsigned int
tcp_print_tuple(char *buffer, const struct ip_nat_tuple *tuple)
{
	return sprintf(buffer, "srcpt=%u dstpt=%u ",
		       ntohs(tuple->src.u.tcp.port),
		       ntohs(tuple->dst.u.tcp.port));
}

static unsigned int
tcp_print_manip(char *buffer, const union ip_nat_proto_manip *prmanip)
{
	return sprintf(buffer, "port=%u ", ntohs(prmanip->tcp.port));
}

/* Print out the private part of the binding (if any). */
static unsigned int tcp_print_private(char *buffer,
				      const struct ip_nat_binding *binding)
{
	struct ip_nat_binding_tcp *binding_tcp
		= ((void *)binding) + proto_tcp.mappingtype.private_offset;
	int len = 0;

	switch (binding_tcp->state) {
	case IP_NAT_TCP_CLOSING:
		len = sprintf(buffer, "CLOSING ");
		break;
	case IP_NAT_TCP_OPENING:
		len = sprintf(buffer, "OPENING ");
		break;
	case IP_NAT_TCP_ESTABLISHED:
		break;
	}

	/* FIXME: if it's handled by a per-proto, print that here. --RR */
	return len;
}

static struct ip_nat_proto proto_tcp
= { NULL, IPPROTO_TCP, "TCP",
    tcp_pkt_to_tuple,
    tcp_manip_pkt,
    tcp_in_range,
    tcp_unique_tuple,
    tcp_invert_tuple,
    tcp_print_tuple,
    tcp_print_manip,
    { NULL, "TCP", 0, &__this_module,
      sizeof(struct ip_nat_binding_tcp), 0,
      &tcp_print_private,
      &tcp_check,
      &tcp_rule,
#ifdef DEBUGNAT
      &tcp_destroy,
#else
      NULL,
#endif
    }
};

static struct nf_setsockopt_ops ip_nat_tcp_sockopts
= { NULL, PF_INET, IP_NAT_TCP_ADD_USER, IP_NAT_TCP_DELETE_USER, do_tcp_ctl };

int init_module(void)
{
	int ret;
	DEBUGP("ip_nat_tcp\n");

	ret = ip_nat_register_proto(&proto_tcp);
	if (ret < 0) return ret;

	ret = nf_register_sockopt(&ip_nat_tcp_sockopts);
	if (ret < 0)
		ip_nat_unregister_proto(&proto_tcp);

	return ret;
}

void cleanup_module()
{
	ip_nat_unregister_proto(&proto_tcp);
	nf_unregister_sockopt(&ip_nat_tcp_sockopts);
}
