#ifndef _IP_STATE_H
#define _IP_STATE_H
/* Connection state tracking for netfilter.  This is separated from,
   but required by, the NAT layer; it can also be used by an iptables
   extention. */

#include <asm/spinlock.h>
#include <linux/types.h>
#include "struct_reserve.h"

/* A `tuple' is a structure containing the information to uniquely
  identify a connection.  ie. if two packets have the same tuple, they
  are in the same connection; if not, they are not. 

  We divide the structure along "manipulatable" and
  "non-manipulatable" lines, for the benefit of the NAT code.
*/

#ifdef DEBUG_STATE
#undef EXPORT_SYMBOL
#define EXPORT_SYMBOL(x)
#define static 
#endif

#ifdef DEBUG_STATE
#define IP_STATE_ASSERT(x)						\
	do {								\
		if (!(x)) printk("ASSERT: %s:%i(%s)\n",			\
				 __FILE__, __LINE__, __FUNCTION__);	\
	} while(0);
#define DEBUGP printk
#else
#define IP_STATE_ASSERT(x)
#define DEBUGP(format, args...)
#endif

/* The protocol-specific manipulable parts of the tuple. */
union ip_state_manip_proto
{
	struct {
		u_int16_t port;
	} tcp;
	struct {
		u_int16_t port;
	} udp;
	struct {
		u_int16_t id;
	} icmp;
	/* Add other protocols here. */

	u_int16_t all;
};

/* The manipulable part of the tuple. */
struct ip_state_manip
{
	u_int32_t ip;
	union ip_state_manip_proto u;
	u_int16_t pad; /* Must be set to 0 for memcmp. */
};

/* This contains the information to distinguish a connection. */
struct ip_state_tuple
{
	struct ip_state_manip src;

	/* These are the parts of the tuple which are fixed. */
	struct {
		u_int32_t ip;
		union {
			struct {
				u_int16_t port;
			} tcp;
			struct {
				u_int16_t port;
			} udp;
			struct {
				u_int8_t type, code;
			} icmp;
			/* Add other protocols here. */

			u_int16_t all;
		} u;

		/* The protocol. */
		u_int16_t protonum;
	} dst;
};


enum ip_statetype
{
	/* Started a new connection to track. */
	IPS_NEW = 32,
	/* Not part of an established, and can't be used to make a new
	   one; eg. bogus ICMP error.  State pointer invalid. */
	IPS_INVALID,
	/* We've seen it before, but not a reply. */
	IPS_NEW_REPEAT,
	/* Part of an established connection. */
	IPS_ESTABLISHED,
	/* Reply on an established connection. */
	IPS_ESTABLISHED_REPLY,
	/* Like NEW, but related to an existing connection. */
	IPS_RELATED,
	/* Like NEW_REPEAT, but related to an existing connection. */
	IPS_RELATED_REPEAT,
	/* Specifies a related connection: in established conn. */
	IPS_ESTABLISHED_SPECIFIES_RELATED,
	/* Specifies a related connection: in established conn (reply). */
	IPS_ESTABLISHED_REPLY_SPECIFIES_RELATED
};

/* You need this to examine/alter ip_state structures and hash. */
extern spinlock_t ip_state_lock;

/* Play with dynamic later; meanwhile if it's good enough for DaveM's
   TCP code... */
#define IP_STATE_HTABLE_SIZE 64

extern struct ip_state_tuple_hash *ip_state_hash[IP_STATE_HTABLE_SIZE];

/* Connections usually have two entries in the hash table: one for
   each way (one will have reply == 0, and one reply == 1).  Some
   connections can half-expire; they will then only have one entry. */
struct ip_state_tuple_hash
{
	struct ip_state_tuple_hash *next;

	struct ip_state_tuple tuple;

	/* NULL if it's not hashed. */
	struct ip_state *state;
};

enum ip_state_status {
	/* We haven't seen a reply packet yet. */
	IPS_AWAITING_REPLY,

	/* We've seen packets both ways. */
	IPS_BOTH_WAYS,

	/* It's an expected connection, but we haven't seen a reply
           packet yet. */
	IPS_EXPECTED_AWAITING_REPLY
};

struct ip_state
{
	/* These are my tuples; original and reply */
	struct ip_state_tuple_hash tuplehash[2];

	enum ip_state_status status;

	/* Timer function; deletes us when we go off. */
	struct timer_list timeout;

	/* If we're expecting another connection */
	struct ip_state_expect *expected;
};

/* If you alter the reply tuple, call me. */
extern void ip_state_rehash(struct ip_state *state_modified);

/* Find a connection corresponding to a tuple. */
extern struct ip_state *ip_state_find(const struct ip_state_tuple *tuple);

/* Delete a state. */
extern void ip_state_delete(struct ip_state *state);

/* Expire in this direction.  If expired in both dirs, delete. */
extern void ip_state_delete_maybe(struct ip_state *state, int direction);

/* Add a related connection. */
extern int ip_state_expect_related(const struct ip_state_tuple *tuple,
				   struct ip_state *related_to);

EXTERN_DECLARE_STRUCT_RESERVABLE(ip_state);
#endif /* _IP_STATE_H */
